@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2019 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134, San Rafael,
-   CA  94903, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
+   CA 94945, U.S.A., +1(415)492-9861, for further information.
 */
 
 /*
@@ -26,9 +26,13 @@
 #include <stdlib.h>
 #include <stdarg.h>
 #include <string.h>
+#include <limits.h>
 
 #include "jbig2.h"
 #include "jbig2_priv.h"
+#include "jbig2_image.h"
+#include "jbig2_page.h"
+#include "jbig2_segment.h"
 
 static void *
 jbig2_default_alloc(Jbig2Allocator *allocator, size_t size)
@@ -57,15 +61,16 @@ static Jbig2Allocator jbig2_default_allocator = {
 void *
 jbig2_alloc(Jbig2Allocator *allocator, size_t size, size_t num)
 {
-    /* check for integer multiplication overflow */
-    if (num > 0 && size >= (size_t) - 0x100 / num)
+    /* Check for integer multiplication overflow when computing
+    the full size of the allocation. */
+    if (num > 0 && size > SIZE_MAX / num)
         return NULL;
     return allocator->alloc(allocator, size * num);
 }
 
 /* jbig2_free and jbig2_realloc moved to the bottom of this file */
 
-static int
+static void
 jbig2_default_error(void *data, const char *msg, Jbig2Severity severity, int32_t seg_idx)
 {
     /* report only fatal errors by default */
@@ -76,8 +81,6 @@ jbig2_default_error(void *data, const char *msg, Jbig2Severity severity, int32_t
         fprintf(stderr, "\n");
         fflush(stderr);
     }
-
-    return 0;
 }
 
 int
@@ -86,24 +89,30 @@ jbig2_error(Jbig2Ctx *ctx, Jbig2Severity severity, int32_t segment_number, const
     char buf[1024];
     va_list ap;
     int n;
-    int code;
 
     va_start(ap, fmt);
     n = vsnprintf(buf, sizeof(buf), fmt, ap);
     va_end(ap);
     if (n < 0 || n == sizeof(buf))
-        strncpy(buf, "jbig2_error: error in generating error string", sizeof(buf));
-    code = ctx->error_callback(ctx->error_callback_data, buf, severity, segment_number);
-    if (severity == JBIG2_SEVERITY_FATAL)
-        code = -1;
-    return code;
+        strncpy(buf, "failed to generate error string", sizeof(buf));
+    ctx->error_callback(ctx->error_callback_data, buf, severity, segment_number);
+    return -1;
 }
 
 Jbig2Ctx *
-jbig2_ctx_new(Jbig2Allocator *allocator, Jbig2Options options, Jbig2GlobalCtx *global_ctx, Jbig2ErrorCallback error_callback, void *error_callback_data)
+jbig2_ctx_new_imp(Jbig2Allocator *allocator, Jbig2Options options, Jbig2GlobalCtx *global_ctx, Jbig2ErrorCallback error_callback, void *error_callback_data, int jbig2_version_major, int jbig2_version_minor)
 {
     Jbig2Ctx *result;
 
+    if (jbig2_version_major != JBIG2_VERSION_MAJOR || jbig2_version_minor != JBIG2_VERSION_MINOR) {
+        Jbig2Ctx fakectx;
+        fakectx.error_callback = error_callback;
+        fakectx.error_callback_data = error_callback_data;
+        jbig2_error(&fakectx, JBIG2_SEVERITY_FATAL, -1, "incompatible jbig2dec header (%d.%d) and library (%d.%d) versions",
+            jbig2_version_major, jbig2_version_minor, JBIG2_VERSION_MAJOR, JBIG2_VERSION_MINOR);
+        return NULL;
+    }
+
     if (allocator == NULL)
         allocator = &jbig2_default_allocator;
     if (error_callback == NULL)
@@ -111,8 +120,8 @@ jbig2_ctx_new(Jbig2Allocator *allocator, Jbig2Options options, Jbig2GlobalCtx *g
 
     result = (Jbig2Ctx *) jbig2_alloc(allocator, sizeof(Jbig2Ctx), 1);
     if (result == NULL) {
-        error_callback(error_callback_data, "initial context allocation failed!", JBIG2_SEVERITY_FATAL, -1);
-        return result;
+        error_callback(error_callback_data, "failed to allocate initial context", JBIG2_SEVERITY_FATAL, -1);
+        return NULL;
     }
 
     result->allocator = allocator;
@@ -129,9 +138,9 @@ jbig2_ctx_new(Jbig2Allocator *allocator, Jbig2Options options, Jbig2GlobalCtx *g
     result->n_segments_max = 16;
     result->segments = jbig2_new(result, Jbig2Segment *, result->n_segments_max);
     if (result->segments == NULL) {
-        error_callback(error_callback_data, "initial segments allocation failed!", JBIG2_SEVERITY_FATAL, -1);
+        error_callback(error_callback_data, "failed to allocate initial segments", JBIG2_SEVERITY_FATAL, -1);
         jbig2_free(allocator, result);
-        return result;
+        return NULL;
     }
     result->segment_index = 0;
 
@@ -139,10 +148,10 @@ jbig2_ctx_new(Jbig2Allocator *allocator, Jbig2Options options, Jbig2GlobalCtx *g
     result->max_page_index = 4;
     result->pages = jbig2_new(result, Jbig2Page, result->max_page_index);
     if (result->pages == NULL) {
-        error_callback(error_callback_data, "initial pages allocation failed!", JBIG2_SEVERITY_FATAL, -1);
+        error_callback(error_callback_data, "failed to allocated initial pages", JBIG2_SEVERITY_FATAL, -1);
         jbig2_free(allocator, result->segments);
         jbig2_free(allocator, result);
-        return result;
+        return NULL;
     }
     {
         int index;
@@ -150,6 +159,14 @@ jbig2_ctx_new(Jbig2Allocator *allocator, Jbig2Options options, Jbig2GlobalCtx *g
         for (index = 0; index < result->max_page_index; index++) {
             result->pages[index].state = JBIG2_PAGE_FREE;
             result->pages[index].number = 0;
+            result->pages[index].width = 0;
+            result->pages[index].height = 0xffffffff;
+            result->pages[index].x_resolution = 0;
+            result->pages[index].y_resolution = 0;
+            result->pages[index].stripe_size = 0;
+            result->pages[index].striped = 0;
+            result->pages[index].end_row = 0;
+            result->pages[index].flags = 0;
             result->pages[index].image = NULL;
         }
     }
@@ -158,9 +175,9 @@ jbig2_ctx_new(Jbig2Allocator *allocator, Jbig2Options options, Jbig2GlobalCtx *g
 }
 
 #define get_uint16(bptr)\
-  (((bptr)[0] << 8) | (bptr)[1])
+    (((bptr)[0] << 8) | (bptr)[1])
 #define get_int16(bptr)\
-  (((int)get_uint16(bptr) ^ 0x8000) - 0x8000)
+    (((int)get_uint16(bptr) ^ 0x8000) - 0x8000)
 
 int16_t
 jbig2_get_int16(const byte *bptr)
@@ -196,8 +213,7 @@ jbig2_get_uint32(const byte *bptr)
  * to (continue to) parse it as part of a jbig2 data stream.
  *
  * Return code: 0 on success
- *             -1 if there is a parsing error, or whatever
- *                the error handling callback returns
+ *             -1 if there is a parsing error
  **/
 int
 jbig2_data_in(Jbig2Ctx *ctx, const unsigned char *data, size_t size)
@@ -212,7 +228,7 @@ jbig2_data_in(Jbig2Ctx *ctx, const unsigned char *data, size_t size)
         while (buf_size < size);
         ctx->buf = jbig2_new(ctx, byte, buf_size);
         if (ctx->buf == NULL) {
-            return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "failed to allocate ctx->buf in jbig2_data_in");
+            return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "failed to allocate buffer when reading data");
         }
         ctx->buf_size = buf_size;
         ctx->buf_rd_ix = 0;
@@ -229,7 +245,7 @@ jbig2_data_in(Jbig2Ctx *ctx, const unsigned char *data, size_t size)
             while (buf_size < ctx->buf_wr_ix - ctx->buf_rd_ix + size);
             buf = jbig2_new(ctx, byte, buf_size);
             if (buf == NULL) {
-                return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "failed to allocate buf in jbig2_data_in");
+                return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "failed to allocate bigger buffer when reading data");
             }
             memcpy(buf, ctx->buf + ctx->buf_rd_ix, ctx->buf_wr_ix - ctx->buf_rd_ix);
             jbig2_free(ctx->allocator, ctx->buf);
@@ -256,9 +272,15 @@ jbig2_data_in(Jbig2Ctx *ctx, const unsigned char *data, size_t size)
             if (ctx->buf_wr_ix - ctx->buf_rd_ix < 9)
                 return 0;
             if (memcmp(ctx->buf + ctx->buf_rd_ix, jbig2_id_string, 8))
-                return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "Not a JBIG2 file header");
+                return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "not a JBIG2 file header");
             /* D.4.2 */
             ctx->file_header_flags = ctx->buf[ctx->buf_rd_ix + 8];
+            /* Check for T.88 amendment 2 */
+            if (ctx->file_header_flags & 0x04)
+                return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "file header indicates use of 12 adaptive template pixels (NYI)");
+            /* Check for T.88 amendment 3 */
+            if (ctx->file_header_flags & 0x08)
+                return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "file header indicates use of colored region segments (NYI)");
             if (ctx->file_header_flags & 0xFC) {
                 jbig2_error(ctx, JBIG2_SEVERITY_WARNING, -1, "reserved bits (2-7) of file header flags are not zero (0x%02x)", ctx->file_header_flags);
             }
@@ -273,7 +295,6 @@ jbig2_data_in(Jbig2Ctx *ctx, const unsigned char *data, size_t size)
                 else
                     jbig2_error(ctx, JBIG2_SEVERITY_INFO, -1, "file header indicates a %d page document", ctx->n_pages);
             } else {            /* number of pages not known */
-
                 ctx->n_pages = 0;
                 ctx->buf_rd_ix += 9;
             }
@@ -284,18 +305,26 @@ jbig2_data_in(Jbig2Ctx *ctx, const unsigned char *data, size_t size)
             } else {
                 ctx->state = JBIG2_FILE_RANDOM_HEADERS;
                 jbig2_error(ctx, JBIG2_SEVERITY_DEBUG, -1, "file header indicates random-access organization");
-
             }
             break;
         case JBIG2_FILE_SEQUENTIAL_HEADER:
         case JBIG2_FILE_RANDOM_HEADERS:
             segment = jbig2_parse_segment_header(ctx, ctx->buf + ctx->buf_rd_ix, ctx->buf_wr_ix - ctx->buf_rd_ix, &header_size);
             if (segment == NULL)
-                return 0;       /* need more data */
+                return 0; /* need more data */
             ctx->buf_rd_ix += header_size;
 
-            if (ctx->n_segments == ctx->n_segments_max)
-                ctx->segments = jbig2_renew(ctx, ctx->segments, Jbig2Segment *, (ctx->n_segments_max <<= 2));
+            if (ctx->n_segments == ctx->n_segments_max) {
+                Jbig2Segment **segments;
+
+                segments = jbig2_renew(ctx, ctx->segments, Jbig2Segment *, (ctx->n_segments_max <<= 2));
+                if (segments == NULL) {
+                    ctx->state = JBIG2_FILE_EOF;
+                    return jbig2_error(ctx, JBIG2_SEVERITY_FATAL, segment->number, "failed to allocate space for more segments");
+                }
+                ctx->segments = segments;
+            }
+
 
             ctx->segments[ctx->n_segments++] = segment;
             if (ctx->state == JBIG2_FILE_RANDOM_HEADERS) {
@@ -307,37 +336,80 @@ jbig2_data_in(Jbig2Ctx *ctx, const unsigned char *data, size_t size)
         case JBIG2_FILE_SEQUENTIAL_BODY:
         case JBIG2_FILE_RANDOM_BODIES:
             segment = ctx->segments[ctx->segment_index];
-            if (segment->data_length > ctx->buf_wr_ix - ctx->buf_rd_ix)
-                return 0;       /* need more data */
+
+            /* immediate generic regions may have unknown size */
+            if (segment->data_length == 0xffffffff && (segment->flags & 63) == 38) {
+                byte *s, *e, *p;
+                int mmr;
+                byte mmr_marker[2] = { 0x00, 0x00 };
+                byte arith_marker[2] = { 0xff, 0xac };
+                byte *desired_marker;
+
+                s = p = ctx->buf + ctx->buf_rd_ix;
+                e = ctx->buf + ctx->buf_wr_ix;
+
+                if (e - p < 18)
+                        return 0; /* need more data */
+
+                mmr = p[17] & 1;
+                p += 18;
+                desired_marker = mmr ? mmr_marker : arith_marker;
+
+                /* look for two byte marker */
+                if (e - p < 2)
+                    return 0; /* need more data */
+
+                while (p[0] != desired_marker[0] || p[1] != desired_marker[1]) {
+                    p++;
+                    if (e - p < 2)
+                        return 0; /* need more data */
+                }
+                p += 2;
+
+                /* the marker is followed by a four byte row count */
+                if (e - p < 4)
+                        return 0; /* need more data */
+                segment->rows = jbig2_get_uint32(p);
+                p += 4;
+
+                segment->data_length = p - s;
+                jbig2_error(ctx, JBIG2_SEVERITY_INFO, segment->number, "unknown length determined to be %u", segment->data_length);
+            }
+            else if (segment->data_length > ctx->buf_wr_ix - ctx->buf_rd_ix)
+                    return 0; /* need more data */
+
             code = jbig2_parse_segment(ctx, segment, ctx->buf + ctx->buf_rd_ix);
             ctx->buf_rd_ix += segment->data_length;
             ctx->segment_index++;
             if (ctx->state == JBIG2_FILE_RANDOM_BODIES) {
                 if (ctx->segment_index == ctx->n_segments)
                     ctx->state = JBIG2_FILE_EOF;
-            } else {            /* JBIG2_FILE_SEQUENCIAL_BODY */
-
+            } else {            /* JBIG2_FILE_SEQUENTIAL_BODY */
                 ctx->state = JBIG2_FILE_SEQUENTIAL_HEADER;
             }
             if (code < 0) {
                 ctx->state = JBIG2_FILE_EOF;
-                return code;
+                return jbig2_error(ctx, JBIG2_SEVERITY_WARNING, segment->number, "failed to decode; treating as end of file");
             }
             break;
         case JBIG2_FILE_EOF:
             if (ctx->buf_rd_ix == ctx->buf_wr_ix)
                 return 0;
-            return jbig2_error(ctx, JBIG2_SEVERITY_WARNING, -1, "Garbage beyond end of file");
+            return jbig2_error(ctx, JBIG2_SEVERITY_WARNING, -1, "garbage beyond end of file");
         }
     }
 }
 
-void
+Jbig2Allocator *
 jbig2_ctx_free(Jbig2Ctx *ctx)
 {
-    Jbig2Allocator *ca = ctx->allocator;
+    Jbig2Allocator *ca;
     int i;
 
+    if (ctx == NULL)
+        return NULL;
+
+    ca = ctx->allocator;
     jbig2_free(ca, ctx->buf);
     if (ctx->segments != NULL) {
         for (i = 0; i < ctx->n_segments; i++)
@@ -353,6 +425,8 @@ jbig2_ctx_free(Jbig2Ctx *ctx)
     }
 
     jbig2_free(ca, ctx);
+
+    return ca;
 }
 
 Jbig2GlobalCtx *
@@ -361,10 +435,10 @@ jbig2_make_global_ctx(Jbig2Ctx *ctx)
     return (Jbig2GlobalCtx *) ctx;
 }
 
-void
+Jbig2Allocator *
 jbig2_global_ctx_free(Jbig2GlobalCtx *global_ctx)
 {
-    jbig2_ctx_free((Jbig2Ctx *) global_ctx);
+    return jbig2_ctx_free((Jbig2Ctx *) global_ctx);
 }
 
 /* I'm not committed to keeping the word stream interface. It's handy
@@ -379,25 +453,37 @@ typedef struct {
 } Jbig2WordStreamBuf;
 
 static int
-jbig2_word_stream_buf_get_next_word(Jbig2WordStream *self, int offset, uint32_t *word)
+jbig2_word_stream_buf_get_next_word(Jbig2WordStream *self, size_t offset, uint32_t *word)
 {
     Jbig2WordStreamBuf *z = (Jbig2WordStreamBuf *) self;
-    const byte *data = z->data;
-    uint32_t result;
+    uint32_t val = 0;
+    int ret = 0;
 
-    if (offset + 4 < z->size)
-        result = (data[offset] << 24) | (data[offset + 1] << 16) | (data[offset + 2] << 8) | data[offset + 3];
-    else if (offset >= z->size)
+    if (self == NULL || word == NULL)
         return -1;
-    else {
-        int i;
-
-        result = 0;
-        for (i = 0; i < z->size - offset; i++)
-            result |= data[offset + i] << ((3 - i) << 3);
+    if (offset >= z->size) {
+        *word = 0;
+        return 0;
     }
-    *word = result;
-    return 0;
+
+    if (offset < z->size) {
+        val |= z->data[offset] << 24;
+        ret++;
+    }
+    if (offset + 1 < z->size) {
+        val |= z->data[offset + 1] << 16;
+        ret++;
+    }
+    if (offset + 2 < z->size) {
+        val |= z->data[offset + 2] << 8;
+        ret++;
+    }
+    if (offset + 3 < z->size) {
+        val |= z->data[offset + 3];
+        ret++;
+    }
+    *word = val;
+    return ret;
 }
 
 Jbig2WordStream *
@@ -406,7 +492,7 @@ jbig2_word_stream_buf_new(Jbig2Ctx *ctx, const byte *data, size_t size)
     Jbig2WordStreamBuf *result = jbig2_new(ctx, Jbig2WordStreamBuf, 1);
 
     if (result == NULL) {
-        jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "failed to allocate Jbig2WordStreamBuf in jbig2_word_stream_buf_new");
+        jbig2_error(ctx, JBIG2_SEVERITY_FATAL, -1, "failed to allocate word stream");
         return NULL;
     }
 
@@ -442,7 +528,7 @@ void *
 jbig2_realloc(Jbig2Allocator *allocator, void *p, size_t size, size_t num)
 {
     /* check for integer multiplication overflow */
-    if (num > 0 && size >= (size_t) - 0x100 / num)
+    if (num > 0 && size >= SIZE_MAX / num)
         return NULL;
     return allocator->realloc(allocator, p, size * num);
 }
